#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

#include "Gdc2.h"
#include "IFtp.h"
#include "GToken.h"

// #define LOG_LISTINGS

///////////////////////////////////////////////////////////////////
int LookupMonth(char *m)
{
	char *Months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	for (int i=0; i<12; i++)
	{
		if (m AND stricmp(m, Months[i]) == 0)
		{
			return i + 1;
		}
	}
	return 1;
}

///////////////////////////////////////////////////////////////////
class ILogProxy : public GSocket
{
	GSocket *Dest;
	IFtpDataPipe *Pipe;
	Progress *Meter;

public:
	ILogProxy(GSocket *dest, IFtpDataPipe *pipe, Progress *meter)
	{
		Dest = dest;
		Pipe = pipe;
		Meter = meter;
	}

	int Write(char *Data, int Len, int Flags)
	{
		if (Pipe)
		{
			return Pipe->WritePipe(this, Data, Len, Flags);
		}

		return GSocket::Write(Data, Len, Flags);
	}

	int Read(char *Data, int Len, int Flags)
	{
		if (Pipe)
		{
			return Pipe->ReadPipe(this, Data, Len, Flags);
		}

		return GSocket::Read(Data, Len, Flags);
	}

	void OnError(int ErrorCode, char *ErrorDescription)
	{
		if (Dest AND
			ErrorCode > 0 AND
			ErrorDescription)
		{
			char Str[256];
			sprintf(Str, "[Data] %s", ErrorDescription);
			Dest->OnInformation(Str);
		}
	}

	void OnInformation(char *s)
	{
		if (Dest AND s)
		{
			char Str[256];
			sprintf(Str, "[Data] %s", s);
			Dest->OnInformation(Str);
		}
	}

	void OnDisconnect()
	{
		if (Dest)
		{
			Dest->OnInformation("[Data] Disconnect");
		}
	}

	void OnRead(char *Data, int Len)
	{
		if (Meter)
		{
			Meter->Value(Meter->Value() + Len);
		}
	}

	void OnWrite(char *Data, int Len)
	{
		if (Meter)
		{
			Meter->Value(Meter->Value() + Len);
		}
	}
};

///////////////////////////////////////////////////////////////////
IFtpEntry::IFtpEntry()
{
	Attributes = 0;
	Size = 0;
	Name = Date = Time = 0;
}

IFtpEntry::IFtpEntry(IFtpEntry *Entry)
{
	Attributes = Entry->Attributes;
	Size = Entry->Size;
	Name = NewStr(Entry->Name);
	Date = NewStr(Entry->Date);
	Time = NewStr(Entry->Time);
}

IFtpEntry::IFtpEntry(char *Entry)
{
	Attributes = 0;
	Size = 0;
	Name = Date = Time = 0;
	if (Entry)
	{
		char *Ws = " \t";
		TokenParser T(Entry, Ws);
		if (T.Tokens > 7)
		{
			char *n = T.Get(4);
			int SizeElement = (isdigit(*n)) ? 4 : 3;
			n = Entry;
			for (int i=0; i<SizeElement+4; i++)
			{
				for (; *n AND NOT strchr(Ws, *n); n++);
				for (; *n AND strchr(Ws, *n); n++);
			}
			Name = NewStr(n);

			char *f = T.Get(0);
			if (f AND strlen(f) == 10)
			{
				if (f[0] == 'd' OR
					f[0] == 'l') Attributes |= IFTP_DIR;
				if (f[1] != '-') Attributes |= IFTP_READ;
				if (f[2] != '-') Attributes |= IFTP_WRITE;
				if (f[3] != '-') Attributes |= IFTP_EXECUTE;

				if (f[4] != '-') Attributes |= IFTP_GRP_READ;
				if (f[5] != '-') Attributes |= IFTP_GRP_WRITE;
				if (f[6] != '-') Attributes |= IFTP_GRP_EXECUTE;

				if (f[7] != '-') Attributes |= IFTP_GLOB_READ;
				if (f[8] != '-') Attributes |= IFTP_GLOB_WRITE;
				if (f[9] != '-') Attributes |= IFTP_GLOB_EXECUTE;

				if (f[0] == 'l')
				{
					// link...
					Attributes |= IFTP_SYM_LINK;
					
					char *Arrow = strstr(Name, "->");
					if (Arrow)
					{
						Arrow[-1] = 0;
					}
				}
			}

			char *MonthStr = T.Get(SizeElement+1);
			char *YearOrTime = T.Get(SizeElement+3);
			if (MonthStr AND YearOrTime)
			{
				int Day = atoi(T.Get(SizeElement+2));
				int Month = LookupMonth(MonthStr);
				int Year;

				if (strchr(YearOrTime, ':'))
				{
					// It's time
					Time = NewStr(YearOrTime);

					// Default the year
					time_t Time = time(NULL);
					tm *ptm = localtime(&Time);
					if (ptm)
					{
						Year = 1900 + ptm->tm_year;
					}
				}
				else 
				{
					// It's a year
					Year = atoi(YearOrTime);
				}

				char DateStr[64];
				sprintf(DateStr, "%i/%i/%i", Day, Month, Year);
				Date = NewStr(DateStr);
			}

			Size = atoi(T.Get(SizeElement));
		}
	}
}

IFtpEntry::~IFtpEntry()
{
	DeleteArray(Name);
	DeleteArray(Date);
	DeleteArray(Time);
}

///////////////////////////////////////////////////////////////////
#define VERIFY(i, ret) if (i != ret) { Close(); return 0; }
#define VERIFY_RANGE(i, range) if (((i)/100) != range) { Close(); return 0; }

IFtp::IFtp(IFtpDataPipe *pipe)
{
	F = 0;
	Pipe = pipe;
	Authenticated = false;
	Meter = 0;
	Socket = Data = Listen = 0;

	InBuf[0] = 0;
	Ip[0] = 0;
	Port = 0;
	PassiveMode = false;
	ForceActive = false;

	RestorePos = 0;
	AbortTransfer = false;
}

IFtp::~IFtp()
{
}

int IFtp::WriteLine(char *Msg)
{
	int Status = 0;

	if (Socket)
	{
		char *b = Msg ? Msg : OutBuf;
		int l = strlen(b);
		Status = Socket->Write(b, l, 0);
	}

	return Status;	
}

int IFtp::ReadLine(char *Msg)
{
	int Status = 0;

	while (Socket)
	{
		// look through the input list for result codes
		char *s;
		while (s = In.First())
		{
			char Temp[32], *d = Temp;
			char *c;
			for (c=s; isdigit(*c); c++)
			{
				*d++ = *c;
			}
			*d++ = 0;
			if (c[0] == ' ')
			{
				if (Msg)
				{
					strcpy(Msg, c+1);
				}
				Status = atoi(Temp);
			}

			delete [] s;
			In.Delete(s);

			if (Status)
			{
				return Status;
			}
		}

		// Ok, no result code in the input list so
		// Read data off socket
		int Len = strlen(InBuf);
		int Read = 0;

		if (Socket)
		{
			Read = Socket->Read(InBuf+Len, sizeof(InBuf)-Len-1, 0);
		}

		if (Read > 0)
		{
			// process input into lines
			char *Eol = 0;
			InBuf[Len+Read] = 0;
			do
			{
				Eol = strnstr(InBuf, "\r\n", sizeof(InBuf)-((int)s-(int)InBuf));
				if (Eol)
				{
					// copy line into input list
					char *New = 0;
					In.Insert(New = NewStr(InBuf, (int)Eol-(int)InBuf));
					Eol += 2;

					// move data after line up to the start of the buffer
					int EolLen = strlen(Eol);
					LgiAssert(EolLen < 255);
					memmove(InBuf, Eol, EolLen+1);
				}
			}
			while (Eol);
		}
		else break; // the read failed....

	} // loop around to check results

	return Status;
}

bool IFtp::IsOpen()
{
	return false;
}

bool IFtp::Open(GSocket *S, char *RemoteHost, int Port, char *User, char *Password)
{
	Socket = S;
	Authenticated = false;	

	if (NOT Port) Port = 21;

	if (Socket AND
		Socket->Open(RemoteHost, Port))
	{
		VERIFY(ReadLine(), 220);

		if (User AND Password)
		{
			// User/pass
			sprintf(OutBuf, "USER %s\r\n", User);
			if (WriteLine())
			{
				VERIFY(ReadLine(), 331);

				sprintf(OutBuf, "PASS %s\r\n", Password);
				if (WriteLine())
				{
					VERIFY(ReadLine(), 230);
					Authenticated = true;
				}
			}
		}
		else
		{
			// Anonymous
			sprintf(OutBuf, "USER anonymous\r\n");
			if (WriteLine())
			{
				VERIFY(ReadLine(), 331);

				sprintf(OutBuf, "PASS me@privacy.net\r\n"); // garrenteed to never
														// be allocated to an IP
				if (WriteLine())
				{
					VERIFY(ReadLine(), 230);

					Authenticated = true;
				}
			}
		}
	}

	return Authenticated;
}

bool IFtp::Close()
{
	if (F)
	{
		F->Close();
	}

	if (Socket)
	{
		Socket->Close();
		Socket = 0;
	}

	if (Data)
	{
		Data->Close();
		Data = 0;
	}

	if (Listen)
	{
		Listen->Close();
		Listen = 0;
	}

	return true;
}

void IFtp::Noop()
{
	if (Socket)
	{
		sprintf(OutBuf, "NOOP\r\n");
		if (WriteLine())
		{
			ReadLine();
		}
	}
}

bool IFtp::GetDir(char *Dir)
{
	if (Socket AND
		Dir)
	{
		char Temp[256];

		sprintf(OutBuf, "PWD\r\n");
		WriteLine();
		VERIFY(ReadLine(Temp), 257);

		char *Start = strchr(Temp, '\"');
		char *End = (Start) ? strchr(Start+1, '\"') : 0;
		if (End)
		{
			*End = 0;
			strcpy(Dir, Start+1);
		}
		
		return true;
	}

	return false;
}

bool IFtp::SetDir(char *Dir)
{
	if (Socket AND Dir)
	{
		sprintf(OutBuf, "CWD %s\r\n", Dir);
		WriteLine();
		VERIFY(ReadLine(), 250);

		return true;
	}

	return false;
}

bool IFtp::CreateDir(char *Dir)
{
	if (Socket AND Dir)
	{
		sprintf(OutBuf, "MKD %s\r\n", Dir);
		WriteLine();
		VERIFY_RANGE(ReadLine(), 2);

		return true;
	}

	return false;
}

bool IFtp::DeleteDir(char *Dir)
{
	if (Socket AND Dir)
	{
		sprintf(OutBuf, "RMD %s\r\n", Dir);
		WriteLine();
		VERIFY_RANGE(ReadLine(), 2);

		return true;
	}

	return false;
}

bool IFtp::UpDir()
{
	if (Socket)
	{
		sprintf(OutBuf, "CDUP\r\n");
		WriteLine();
		VERIFY(ReadLine(), 250);

		return true;
	}

	return false;
}

bool IFtp::ListDir(List<IFtpEntry> *Dir)
{
	if (Dir AND
		Socket AND
		SetupData())
	{
		BytePipe Buf;

		// List command
		sprintf(OutBuf, "LIST\r\n");
		WriteLine();

		// Accept the data connection
		if (ConnectData())
		{
			VERIFY_RANGE(ReadLine(), 1);

			if (Meter)
			{
				Meter->SetLimits(0, 0);
				Meter->Value(0);
			}

			// Read the data
			uchar Buffer[1024];
			int Len;
			
			while ((Len = Data->Read((char*) Buffer, sizeof(Buffer), 0)) > 0)
			{
				if (Meter)
				{
					Meter->Value(Buf.Sizeof());
				}

				Buf.Push(Buffer, Len);
			}

			if (Meter)
			{
				Meter->Value(0);
			}
		}
		DeleteObj(Data);
		DeleteObj(Listen);

		// Parse the results
		int Len = Buf.Sizeof();
		char *Text = NEW(char[Len+1]);
		if (Text AND Buf.Pop((uchar*)Text, Len))
		{
			Text[Len] = 0;

			#ifdef LOG_LISTINGS
			GFile F;
			if (F.Open("ftplog.txt", O_WRITE))
			{
				F.Write(Text, Len);
				F.Close();
			}
			#endif

			TokenParser Lines(Text, "\r\n");
			for (int i=0; i<Lines.Tokens; i++)
			{
				IFtpEntry *e = NEW(IFtpEntry(Lines.Get(i)));
				if (e AND e->Name)
				{
					if (strcmp(e->Name, ".") != 0 AND
						strcmp(e->Name, "..") != 0)
					{
						Dir->Insert(e);
						e = 0;
					}
				}
				
				DeleteObj(e);
			}
			
			VERIFY_RANGE(ReadLine(), 2);
			return true;
		}
	}

	return false;
}

bool IFtp::RenameFile(char *From, char *To)
{
	if (Socket AND From AND To)
	{
		// From
		sprintf(OutBuf, "RNFR %s\r\n", From);
		WriteLine();
		VERIFY_RANGE(ReadLine(), 3);

		// To
		sprintf(OutBuf, "RNTO %s\r\n", To);
		WriteLine();
		VERIFY_RANGE(ReadLine(), 2);

		return true;
	}
	return false;
}

bool IFtp::DeleteFile(char *Remote)
{
	if (Socket AND Remote)
	{
		sprintf(OutBuf, "DELE %s\r\n", Remote);
		WriteLine();
		VERIFY_RANGE(ReadLine(), 2);
		return true;
	}
	return false;
}

bool IFtp::DownloadFile(char *Local, IFtpEntry *Remote)
{
	if (Remote AND Remote->Name)
	{
		return TransferFile(Local, Remote->Name, Remote->Size, false);
	}
	return false;
}

bool IFtp::UploadFile(char *Local, char *Remote)
{
	return TransferFile(Local, Remote, FileSize(Local), true);
}

bool IFtp::ResumeAt(int Pos)
{
	sprintf(OutBuf, "REST %i\r\n", Pos);
	WriteLine();
	int FtpStatus = ReadLine();
	bool Status = (FtpStatus/100) < 4;
	RestorePos = (Status) ? Pos : 0;
	return Status;
}

bool IFtp::TransferFile(char *Local, char *Remote, int Size, bool Upload)
{
	bool Status = false;
	bool Aborted = false;

	F = NEW(GFile);

	if (F AND
		Socket AND
		Local AND
		Remote)
	{
		if (F->Open(Local, (Upload)?O_READ:O_WRITE))
		{
			if (SetupData())
			{
				// Data command
				sprintf(OutBuf, "%s %s\r\n", (Upload)?"STOR":"RETR", Remote);
				WriteLine();

				// Build data connection
				if (ConnectData())
				{
					VERIFY_RANGE(ReadLine(), 1);

					// do the transfer
					int TempLen = 64 << 10;
					uchar *Temp = NEW(uchar[TempLen]);
					if (Temp)
					{
						int Processed = 0;
						int Len = 0;

						if (Meter)
						{
							Meter->SetLimits(0, Size-1);
						}
						
						if (NOT Upload)
						{
							Size -= RestorePos;
						}
						AbortTransfer = false;

						bool Error = false;
						while (	Socket AND
								Data AND
								Processed < Size AND
								NOT Error AND
								NOT AbortTransfer AND
								(NOT Meter OR NOT Meter->Cancel()))
						{
							if (Meter)
							{
								Meter->Value(0);
							}

							if (Upload)
							{
								// upload loop
								do
								{
									Len = F->Read(Temp, TempLen);
									if (Len > 0)
									{
										int WriteLen = 0;
										
										if (Data)
										{
											WriteLen = Data->Write((char*) Temp, Len, 0);
										}

										if (WriteLen == Len)
										{
											Processed += Len;
										}
										else
										{
											Error = true;
										}
									}

									if (Meter AND Meter->Cancel())
									{
										break;
									}
								}
								while (	Len > 0 AND
										NOT AbortTransfer);
							}
							else
							{
								// restore support
								F->Seek(RestorePos, SEEK_SET);
								if (Meter AND RestorePos > 0)
								{
									Meter->SetParameter(10, RestorePos);
								}
								RestorePos = 0;

								// download loop
								while (Data)
								{
									if (Data)
									{
										Len = Data->Read((char*) Temp, TempLen, 0);
									}

									if (Len <= 0 OR
										AbortTransfer)
									{
										break;
									}

									F->Write(Temp, Len);
									Processed += Len;
									if (Meter)
									{
										if (Meter->Cancel())
										{
											break;
										}
									}
								}

								if (Len <= 0)
								{
									Error = true;
								}
							}
						}

						if (AbortTransfer OR
							(Meter AND Meter->Cancel()))
						{
							// send abort command
							sprintf(OutBuf, "ABOR\r\n");
							WriteLine();
							if ((ReadLine()/100) == 4) // temp msg
							{
								// response... we don't care just read it
								ReadLine();
							}

							Aborted = true;
						}

						if (Data)
						{
							Data->Close();
						}

						Status = Size == Processed;
						if (Meter)
						{
							Meter->Value(0);
						}

						DeleteArray(Temp);
					}

					// read eof response
					if (NOT Aborted)
					{
						VERIFY_RANGE(ReadLine(), 2);
					}
				}

				DeleteObj(Data);
				DeleteObj(Listen);
			}
			else
			{
				// data setup failed
			}
		}
		else
		{
			// couldn't open file...
			char Str[256];
			sprintf(Str, "Error: Couldn't open '%s' for %s", Local, (Upload)?(char*)"reading":(char*)"writing");
			Socket->OnInformation(Str);
		}
	}

	DeleteObj(F);

	return Status;
}

bool IFtp::SetupData()
{
	bool Status = false;

	DeleteObj(Data);

	if (Socket)
	{
		Data = NEW(ILogProxy(Socket, Pipe, Meter));
	}

	if (Data)
	{
		PassiveMode = false;

		// Type command
		sprintf(OutBuf, "TYPE I\r\n");
		if (WriteLine())
		{
			VERIFY_RANGE(ReadLine(), 2);
		}

		if (NOT ForceActive)
		{
			// Try the PASV command first... it's better for
			// getting out of a firewall to the server
			sprintf(OutBuf, "PASV\r\n");
			WriteLine();

			char Temp[256];
			if (ReadLine(Temp)/100 == 2)
			{
				// Ok we have PASV mode
				// grab the IP and Port
				char *Start = strchr(Temp, '(');
				if (Start)
				{
					Start++;
					char *End = strchr(Start, ')');
					if (End)
					{
						*End = 0;
						TokenParser T(Start, ",");
						if (T.Tokens == 6)
						{
							sprintf(Ip, "%s.%s.%s.%s", T.Get(0), T.Get(1), T.Get(2), T.Get(3));
							Port = (atoi(T.Get(4))<<8) | atoi(T.Get(5));
							PassiveMode = true;
							return true;
						}
					}
				}
			}
			else
			{
				return false;
			}
		}

		// Ok, no PASV so we use the PORT command

		// Port command
		// static int p = (4<<8) | 27;

		bool ListenStatus = false;
		Listen = NEW(GSocket);			// listen
		if (Listen)
		{
			if (Listen->Listen())
			{
				Socket->WhatsMyIp(Ip);				// get current IP
				Port = Listen->WhatsMyPort();		// get current port

				ListenStatus = true;
			}
		}

		if (ListenStatus)
		{
			// convert IP to PORT's comma delimited format
			for (char *p = Ip; *p; p++) if (*p == '.') *p = ',';
			sprintf(OutBuf, "PORT %s,%i,%i\r\n", Ip, Port>>8, Port&0xFF);
			WriteLine();
			VERIFY_RANGE(ReadLine(), 2);
			return true;
		}
	}

	return false;
}

bool IFtp::ConnectData()
{
	if (PassiveMode)
	{
		return Data->Open(Ip, Port);
	}
	else
	{
		Data = NEW(ILogProxy(Socket, Pipe, Meter));
		if (Data)
		{
			return Listen->Accept(Data);
		}
	}

	return false;
}

bool IFtp::SetPerms(char *File, int Perms)
{
	if (Socket AND File)
	{
		sprintf(OutBuf, "SITE CHMOD %03.3X %s\r\n", Perms, File);
		WriteLine();
		VERIFY_RANGE(ReadLine(), 2);

		return true;
	}

	return false;
}
