/*hdr
**      FILE:           Mail.h
**      AUTHOR:         Matthew Allen
**      DATE:           28/5/98
**      DESCRIPTION:    Mail app header
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#ifndef __MAIL_H
#define __MAIL_H

#include "INet.h"
#include "Base64.h"
#include "Progress.h"

#define SMTP_PORT						25
#define POP3_PORT						110
#define MAX_LINE_SIZE					1024
#define MAX_NAME_SIZE					64
#define EMAIL_LINE_SIZE					76

// MIME content types
#define CONTENT_NONE					0
#define CONTENT_BASE64					1
#define CONTENT_QUOTED_PRINTABLE		2
#define CONTENT_OCTET_STREAM			3

// Mail logging defines
#define MAIL_SEND_LINE					0
#define MAIL_RECEIVE_LINE				1
#define MAIL_ERROR_LINE					2
#define MAIL_WARNING_LINE				3

#define MAIL_SEND_COLOUR				Rgb24(0, 0, 0xff)
#define MAIL_RECEIVE_COLOUR				Rgb24(0, 0x8f, 0)
#define MAIL_ERROR_COLOUR				Rgb24(0xff, 0, 0)
#define MAIL_WARNING_COLOUR				Rgb24(0xff, 0x7f, 0)

// Helper functions
extern void TokeniseStrList(char *Str, List<char> &Output, char *Delim);
extern char ConvHexToBin(char c);
#define ConvBinToHex(i) (((i)<10)?'0'+(i):'A'+(i)-10)
extern void DecodeAddrName(char *Start, char *&Name, char *&Addr, char *DefaultDomain);
extern char *DecodeRfc2047(char *Str);
extern char *EncodeRfc2047(char *Str);
extern char *DecodeBase64Str(char *Str, int Len = -1);
extern char *DecodeQuotedPrintableStr(char *Str, int Len = -1);
extern bool Is8Bit(char *Text);
extern int MaxLineLen(char *Text);

// Classes
class MailProtocol;

class LogEntry
{
public:
	char *Text;
	int Data;
	COLOUR c;

	LogEntry(char *text, int data = 0, COLOUR col = 0)
	{
		Text = NewStr(text);
		Data = data;
		c = col;
	}

	~LogEntry()
	{
		DeleteArray(Text);
	}
};

class FileDescriptor : public GObject
{
protected:
	// Global
	int Size;
	char *MimeType;

	// Read from file
	GFile File;
	GFile *Embeded;
	int Offset;

	// Write to memory
	uchar *Data;

public:
	FileDescriptor(GFile *embed, int Offset, int Size, char *Name);
	FileDescriptor(char *name);
	FileDescriptor();
	~FileDescriptor();

	// Access functions
	GFile *GotoObject();		// Get data to read
	uchar *GetData();			// Get data from write
	int Sizeof();
	char *GetMimeType() { return MimeType; }
	void SetMimeType(char *s) { DeleteArray(MimeType); MimeType = NewStr(s); }

	// Decode MIME data to memory
	bool Decode(char *ContentType,
				char *ContentTransferEncoding,
				char *MimeData,
				int MimeDataLength);
};

class AddressDescriptor : public GObject {
protected:
public:
	bool Status;
	uchar CC;
	char *Name;
	char *Addr;

	AddressDescriptor(AddressDescriptor *Copy = 0);
	~AddressDescriptor();

	void Print(char *Str);
	int Sizeof();
	bool Serialize(GFile &f, bool Write);
};

// FIELD_PRIORITY is equivilent to the header field: X-Priority
//	1 - High
//	...
//	5 - Low
#define MAIL_PRIORITY_HIGH				1
#define MAIL_PRIORITY_NORMAL			3 // ???
#define MAIL_PRIORITY_LOW				5

class MailMessage {
public:
	List<AddressDescriptor>		To;
	AddressDescriptor			From;
	AddressDescriptor			Reply;
	char*						Subject;
	char*						Text;
	char*						MessageID;
	List<FileDescriptor>		Attachments;
	char*						InternetHeader;
	char						Priority;
	int16						CodePage;
	char*						HtmlText;

	MailMessage();
	virtual ~MailMessage();
	void Empty();
	
	// Conversion to/from MIME
	virtual bool Encode(StringPipe &Out, MailProtocol *Protocol, bool Mime = true);
	virtual bool EncodeHeaders(StringPipe &Out, MailProtocol *Protocol, bool Mime = true);
	virtual bool EncodeBody(StringPipe &Out, MailProtocol *Protocol, bool Mime = true);
	void EncodeQuotedPrintable(StringPipe &Out, char *In);
	void EncodeText(StringPipe &Out, char *In);
};

class MailProtocol
{
protected:
	char Buffer[4<<10];
	GSocket *Socket;

	bool Read();
	bool Write(char *Buf = NULL, bool Log = false);
	void LogStr(char *Str);
	virtual void OnUserMessage(char *Str) {}

public:
	GSemaphore *Lock;
	Progress *Info;
	List<LogEntry> *Log;

	char *ServerMsg;
	char *ProgramName;
	char *DefaultDomain;

	MailProtocol();
	virtual ~MailProtocol();

	bool _Lock();
	void _Unlock();
	bool _Cancel();
	GSocket *GetSocket() { return Socket; }
};

class MailSmtp : public MailProtocol {
protected:
	bool ReadReply(char *Str, char *Line = 0);
	bool WriteText(char *Str);

public:
	MailSmtp();
	~MailSmtp();

	bool Open(GSocket *S, char *RemoteHost, char *UserName, char *Password, int Port = SMTP_PORT);
	bool Send(MailMessage *Msg, bool Mime = FALSE);
	bool Close();
};

typedef List<MailMessage> MailList;

class MailSource : public MailProtocol {
public:
	virtual ~MailSource() {}

	// Connection
	virtual bool Open(	GSocket *S,
						char *RemoteHost,
						int Port,
						char *User,
						char *Password) = 0;
	virtual bool Close() = 0;

	// Commands available while connected
	virtual int GetMessages() = 0;
	virtual bool Receive(int Message, MailMessage *Msg, bool Delete = true, int MaxSize = -1) = 0;
	virtual bool Delete(int Message) = 0;
	virtual int Sizeof(int Message) = 0;
	virtual bool GetUid(int Message, char *Id) = 0;
	virtual bool GetUidList(List<char> &Id) = 0;
	virtual char *GetHeaders(int Message) = 0;
};

class MailPop3 : public MailSource {
protected:
	bool ReadReply();
	bool ReadMultiLineReply(char *&Str);
	int GetInt();
	bool MailIsEnd(char *Ptr, int Len);

	char *End;
	char *Marker;
	int Messages;

public:
	MailPop3();
	~MailPop3();

	// Connection
	bool Open(GSocket *S, char *RemoteHost, int Port, char *User, char *Password);
	bool Close();

	// Commands available while connected
	int GetMessages() { return Messages; }
	bool Receive(int Message, MailMessage *Msg, bool Delete = true, int MaxSize = -1);
	bool Delete(int Message);
	int Sizeof(int Message);
	bool GetUid(int Message, char *Id);
	bool GetUidList(List<char> &Id);
	char *GetHeaders(int Message);
};

class MailIMap : public MailSource {
protected:
	GSocket *S;
	int NextCmd;
	char Buf[256];
	List<char> Dialog;
	List<char> Uid;

	void ClearDialog();
	void ClearUid();
	bool FillUidList();
	bool WriteBuf();
	bool ReadResponse(int Cmd = -1);
	bool GetParts(int Message, char *&Ptr, char *Parts);

public:
	MailIMap();
	~MailIMap();

	// Connection
	bool Open(GSocket *S, char *RemoteHost, int Port, char *User, char *Password);
	bool Close();

	// Commands available while connected
	int GetMessages();
	bool Receive(int Message, MailMessage *Msg, bool Delete = true, int MaxSize = -1);
	bool Delete(int Message);
	int Sizeof(int Message);
	bool GetUid(int Message, char *Id);
	bool GetUidList(List<char> &Id);
	char *GetHeaders(int Message);
};

#endif