/*hdr
**      FILE:           Store.h
**      AUTHOR:         Matthew Allen
**      DATE:           29/10/98
**      DESCRIPTION:    Tree storage
**
**      Copyright (C) 1998 Matthew Allen
**              fret@ozemail.com.au
*/

#ifndef __STORE2_H
#define __STORE2_H

#include "Gdc2.h"
#include "Contain.h"
#include "Progress.h"
#include "StoreCommon.h"

class WriterItem;

namespace Storage2
{
	#define STORAGE2_ITEM_MAGIC		0x123400AB
	#define STORAGE2_MAGIC			0x123400FE

	class StorageKitImpl;
	class StorageItemImpl;

	struct StorageItemHeader
	{
		int Magic;			// sanity check, always == STORAGE2_ITEM_MAGIC

		// object descriptor
		int Type;			// application defined object type
		int DataLoc;		// location in file of this objects data segment
		int DataSize;		// size of this objects data segment
		
		// heirarchy descriptor
		int ParentLoc;		// file location of objects parent
		int DirLoc;			// file location of the child directory
		int DirCount;		// number of elements in the child directory array
		int DirAlloc;		// numder of allocated elements in child directory

		// file io
		StorageItemHeader();
		bool Serialize(GFile &f, bool Write);
	};

	class StorageItemImpl : public StorageItem
	{
		friend class StorageKitImpl;
		friend class StorageObj;
		
		friend class WriterItem;

	private:
		int StoreLoc;
		bool HeaderDirty;
		StorageItemHeader *Header;		// pointer to our block in the parents Dir array
		StorageItemHeader *Dir;			// an array of children
		StorageItemHeader *Temp;		// we own this block of memory

		StorageItemImpl *Next;
		StorageItemImpl *Prev;
		StorageItemImpl *Parent;
		StorageItemImpl *Child;

		StorageKitImpl *Tree;

		bool IncDirSize();
		bool DirChange();

		void SetDirty(bool Dirty);
		bool SerializeHeader(GFile &f, bool Write);
		bool SerializeObject(GFile &f, bool Write);

	public:
		StorageItemImpl(StorageItemHeader *header);
		~StorageItemImpl();

		// Heirarchy
		StorageItem *GetNext();
		StorageItem *GetPrev();
		StorageItem *GetParent();
		StorageItem *GetChild();

		StorageItem *CreateNext(StorageObj *Obj);
		StorageItem *CreateChild(StorageObj *Obj);
		StorageItem *CreateSub(StorageObj *Obj);
		bool InsertSub(StorageItem *Obj);
		bool DeleteChild(StorageItem *Obj);
		bool DeleteAllChildren();

		// Properties
		int GetType();
		void SetType(int i);
		int GetTotalSize();
		int GetObjectSize();
		StorageKit *GetTree();

		// Impl
		GFile *GotoObject();
		bool EndOfObj(GFile &f);
		bool Save();
	};

	class Block {
	public:
		StorageItemImpl *Dir;
		StorageItemImpl *Object;
		int Start, Length;
	};

	struct StorageHeader {

		int Magic;
		int Version;
		int Reserved2;
		int FirstLoc;
		char Password[32];
		int Reserved3[4];
	};

	class StorageKitImpl :
		public GObject,
		public StorageKit,
		public GSemaphore
	{
		friend class StorageItemImpl;

	protected:
		// File stuff
		char *FileName;
		GFile File;
		bool Status;
		int StoreLoc;
		int Version;
		bool ReadOnly;
		GPassword Password;

		// Root stuff
		StorageItemHeader RootHeader;
		StorageItemImpl *Root;

		// Compact stuff
		uint UserData;
		ValidateItemFunc Func;
		Progress *_Ui;
		DoEvery _Timer;
		int _CompactPos;

		// Member funcs
		void _AddItem(	StorageItemImpl *Item,
						List<Block> &Blocks);
		bool _SortItem(	Block *Item,
						List<Block> &Blocks);
		bool _ValidLoc(int Loc) { return Loc >= 64 AND Loc <= File.GetSize(); }
		bool _Serialize(GFile &f, bool Write);

	public:
		StorageKitImpl(char *FileName);
		~StorageKitImpl();

		int GetStatus() { return Status; }
		int GetVersion() { return Version; }
		quad GetFileSize();
		bool GetPassword(GPassword *p);
		bool SetPassword(GPassword *p);
		
		StorageItem *GetRoot();
		StorageItem *CreateRoot(StorageObj *Obj);

		bool Compact(Progress *p, ValidateItemFunc func = 0, uint userdata = 0);
		bool DeleteItem(StorageItem *Item);
		bool SeparateItem(StorageItem *Item);
		bool ReparentItem(StorageItem *Item, StorageItem *NewParent, bool Next = false);
	};

}

#endif