/*
**	FILE:			File.h
**	AUTHOR:			Matthew Allen
**	DATE:			11/7/95
**	DESCRIPTION:	File subsystem header
**
**
**	Copyright (C) 1995, Matthew Allen
**		fret@ozemail.com.au
*/

#ifndef __FILE_H
#define __FILE_H

#include <dos.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <fcntl.h>
#include <io.h>

#include "LibDefs.h"
#include "LgiOsDefs.h"
#include "VMem.h"

#define MAX_PATH						260

#define INVALID_HANDLE					INVALID_HANDLE_VALUE
#define O_READ							GENERIC_READ
#define O_WRITE							GENERIC_WRITE
#define O_READWRITE						(GENERIC_READ | GENERIC_WRITE)
#define ValidHandle(hnd)				((hnd) != INVALID_HANDLE_VALUE)

#define OPEN_ACCESS_READONLY			(0x0000)
#define OPEN_ACCESS_WRITEONLY			(0x0001)
#define OPEN_ACCESS_READWRITE			(0x0002)
#define OPEN_ACCESS_RO_NOMODLASTACCESS	(0x0004)

// Time/Date related functions infomation
#define DECNANOPERSEC					10000000
#define SECPERMINUTE					60
#define SECPERHOUR						(60 * SECPERMINUTE)
#define SECPERDAY						(24 * SECPERHOUR)

// Disk types
#define DT_3_5FLOPPY					0
#define DT_5_25FLOPPY					1
#define DT_HARDDISK						2
#define DT_CDROM						3
#define DT_NET							4
#define DT_RAM							5
#define DT_VIRTUAL						6
#define DT_MAX							7

// File attributes
#define FA_NORMAL						0x0000
#define FA_READONLY						0x0001
#define FA_HIDDEN						0x0002
#define FA_SYSTEM						0x0004
#define FA_VOLUME						0x0008
#define FA_DIRECTORY					0x0010
#define FA_ARCHIVE						0x0020

class LgiClass VData
{
public:
	virtual ~VData() {}

	// Init/Shutdown
	virtual int Open(char *Str = 0, int Int = 0) = 0;
	virtual int Close() = 0;

	// Properties
	virtual char *GetName() = 0;
	virtual bool SetName(char *Name) = 0;
	virtual int GetSize() = 0;
	virtual int SetSize(int Size) = 0;
	virtual int GetPos() = 0;
	virtual int SetPos(int Pos) = 0;

	// Function
	virtual int Read(void *Buffer, int Size, int Flags = 0) = 0;
	virtual int Write(void *Buffer, int Size, int Flags = 0) = 0;
};

class LgiClass NodeInfo {
protected:
	char *Name;
	int Attributes;
	quad Size;

public:
	NodeInfo();
	NodeInfo(char *n, int a, quad s);
	virtual ~NodeInfo();

	virtual int operator ^(NodeInfo &N);
	char *GetName() { return Name; }
	int GetAttributes() { return Attributes; }
	quad GetSize() { return Size; }
};

class LgiClass VDirectory {
public:
	VDirectory() { }
	virtual ~VDirectory() { Close(); }
	virtual void GetDirType(char *Str, int Len) {}
	virtual int First(char *Name, char Must, char Search) { return TRUE; }
	virtual int Next() { return TRUE; }
	virtual int Close() { return TRUE; }

	virtual long GetAttributes() { return 0; }
	virtual bool IsDir() { return false; }
	virtual char *GetName() { return NULL; }
	virtual void GetName(char *Buf, int Len) {}
	virtual const quad GetCreationTime() { return 0; }
	virtual const quad GetLastAccessTime() { return 0; }
	virtual const quad GetLastWriteTime() { return 0; }
	virtual const quad GetSize() { return 0; }

	bool ConvertToTime(char *Str, quad Time);
	bool ConvertToDate(char *Str, quad Time);
};

class LgiClass VDirView : public VDirectory {

	int Items;
	int Files;
	int Dirs;
	char CurrentDir[260];

	class Node : public NodeInfo {

		Node *Left, *Right;

	public:
		Node();
		Node(char *n, int a, quad s);
		~Node();

		Node **Traverse(Node **ppNode, int Type);
		void Add(Node *Info);

	} *Root, **Index;

public:
	VDirView();
	~VDirView();

	bool Read(char *Dir);

	int GetFiles() { return Files; }
	NodeInfo *File(int i)
	{
		if (Index) return (NodeInfo*) Index[Dirs + i];
		return 0;
	}

	int GetDirs() { return Dirs; }
	NodeInfo *Dir(int i)
	{
		if (Index) return (NodeInfo*) Index[i];
		return 0;
	}

	int GetItems() { return Items; }
	NodeInfo *operator [](int i)
	{
		if (Index) return (NodeInfo*) Index[i];
		return 0;
	}
};

class LgiClass VWin32Dir : public VDirView {

	bool		Open;
	HANDLE		Handle;
	WIN32_FIND_DATA	Data;

public:
	VWin32Dir();
	~VWin32Dir();

	int First(char *Name, char Must, char Search);
	int Next();
	int Close();

	long GetAttributes();
	bool IsDir();
	char *GetName();
	void GetName(char *Buf, int Len);
	const quad GetCreationTime();
	const quad GetLastAccessTime();
	const quad GetLastWriteTime();
	const quad GetSize();
};

// classes
#define VIF_CASE_SENSITIVE			0x0001
#define VIF_CASE_PRESERVED			0x0002
#define VIF_UNICODE_ON_DISK			0x0004
#define VIF_LFN_API				0x4000
#define VIF_COMPRESSED				0x8000

class LgiClass VolumeInfo {
public:
	char		Drive;
	char		Name[32];
	char		System[32];
	ulong		MaxPath;
	ulong		Flags;
};

#define DIF_NONE				0
#define DIF_3_5FLOPPY				1
#define DIF_5_25FLOPPY				2
#define DIF_HARDDISK				3
#define DIF_CDROM				4
#define DIF_NET					5
#define DIF_RAM					6
#define DIF_REMOVABLE				7
#define DIF_MAX					9

class LgiClass DriveInfo {
public:
	char	Name[16];
	int		Type;
	long	TotalSpace;
	long	FreeSpace;
};

#define FileDev			(FileDevice::GetInstance())

class LgiClass FileDevice
{
	friend class GFile;

	int NumDrive;
	DriveInfo *DriveList;
	int FillDiskList();

	static FileDevice *Instance;

public:
	FileDevice();
	~FileDevice();
	static FileDevice *GetInstance() { return Instance; }

	int GetNumDrives();
	DriveInfo *GetDriveInfo(int Drive);
	VDirectory *GetDir(char *DirName);
	HANDLE OpenFile(char *Name, int Mode, int Attrib);

	bool DeleteFile(char *FileName);
	bool CreateDirectory(char *PathName);
	bool RemoveDirectory(char *PathName);
	bool SetCurrentDirectory(char *PathName);
	bool GetCurrentDirectory(char *PathName, int Length);
	bool MoveFile(char *OldName, char *NewName);
	bool GetVolumeInfomation(char Drive, VolumeInfo *pInfo);
};

class LgiClass GFile {
protected:
	HANDLE	hFile;
	char	*Name;
	int	Attributes;
	bool	Swap;
	bool	Status;

	int SwapRead(uchar *Buf, int Size);
	int SwapWrite(uchar *Buf, int Size);

public:
	GFile();
	virtual ~GFile();

	HANDLE Handle() { return hFile; }

	virtual bool Open(char *Name, int Attrib);
	virtual bool Open() { return ValidHandle(hFile); }
	virtual bool Close();

	virtual int Read(void *Buffer, int Size);
	virtual int Write(void *Buffer, int Size);
	virtual int Seek(int To, int Whence);
	virtual int GetPosition();
	virtual int GetSize();
	virtual bool SetSize(int Size);
	virtual bool Eof();
	void SetStatus(bool s = FALSE);
	bool GetStatus();

	// swap bytes on variable io
	void SetSwap(bool s) { Swap = s; }
	bool GetSwap() { return Swap; }

	// String
	virtual int ReadStr(char *Buf, int Size);
	virtual int WriteStr(char *Buf, int Size);
	virtual int Print(char *Format, ...);

	// Read
	virtual GFile &operator >> (char		&i);
	virtual GFile &operator >> (signed char		&i);
	virtual GFile &operator >> (unsigned char	&i);
	virtual GFile &operator >> (signed short	&i);
	virtual GFile &operator >> (unsigned short	&i);
	virtual GFile &operator >> (signed int		&i);
	virtual GFile &operator >> (unsigned int	&i);
	virtual GFile &operator >> (signed long		&i);
	virtual GFile &operator >> (unsigned long	&i);
	virtual GFile &operator >> (float		&i);
	virtual GFile &operator >> (double		&i);
	virtual GFile &operator >> (long double		&i);

	// Write
	virtual GFile &operator << (char		i);
	virtual GFile &operator << (signed char		i);
	virtual GFile &operator << (unsigned char	i);
	virtual GFile &operator << (signed short	i);
	virtual GFile &operator << (unsigned short	i);
	virtual GFile &operator << (signed int		i);
	virtual GFile &operator << (unsigned int	i);
	virtual GFile &operator << (signed long		i);
	virtual GFile &operator << (unsigned long	i);
	virtual GFile &operator << (float		i);
	virtual GFile &operator << (double		i);
	virtual GFile &operator << (long double		i);
};

class LgiClass VFileBuf : public GFile {

	int BufPos;
	int BufSize;
	uchar *Data;

	int SwapRead(uchar *Buf, int Size);
	int SwapWrite(uchar *Buf, int Size);

public:
	VFileBuf();
	virtual ~VFileBuf();

	virtual bool Open(char *Name, int Attrib);
	virtual bool Open();
	virtual bool Close();

	virtual int Read(void *Buffer, int Size);
	virtual int Write(void *Buffer, int Size);
	virtual int Seek(int To, int Whence);
	virtual int GetPosition();
	virtual int GetSize();
	virtual bool SetSize(int Size);
	virtual bool Eof();
	int ReadStr(char *Buf, int Size);
	int WriteStr(char *Buf, int Size);

	GFile &operator >> (           char & i );
	GFile &operator >> (    signed char & i );
	GFile &operator >> (  unsigned char & i );
	GFile &operator >> (   signed short & i );
	GFile &operator >> ( unsigned short & i );
	GFile &operator >> (     signed int & i );
	GFile &operator >> (   unsigned int & i );
	GFile &operator >> (    signed long & i );
	GFile &operator >> (  unsigned long & i );
	GFile &operator >> (          float & i );
	GFile &operator >> (         double & i );
	GFile &operator >> (    long double & i );

	GFile &operator << (                char  i );
	GFile &operator << (         signed char  i );
	GFile &operator << (       unsigned char  i );
	GFile &operator << (        signed short  i );
	GFile &operator << (      unsigned short  i );
	GFile &operator << (          signed int  i );
	GFile &operator << (        unsigned int  i );
	GFile &operator << (         signed long  i );
	GFile &operator << (       unsigned long  i );
	GFile &operator << (               float  i );
	GFile &operator << (              double  i );
	GFile &operator << (         long double  i );
};

// Functions
LgiFunc bool Match(char *Name, char *Mask);
LgiFunc int FileSize(char *FileName);
LgiFunc bool FileExists(char *File);
LgiFunc bool DirExists(char *Dir);
LgiFunc bool ResolveShortcut(char *LinkFile, char *Path, int Len);
LgiFunc void WriteStr(GFile &f, char *s);
LgiFunc char *ReadStr(GFile &f);
LgiFunc int SizeofStr(char *s);

#endif
